"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.subscribe = void 0;
const createSpawn_1 = require("./createSpawn");
const generateShortId_1 = require("./generateShortId");
const Logger_1 = require("./Logger");
const node_path_1 = __importDefault(require("node:path"));
const p_retry_1 = __importDefault(require("p-retry"));
const throttle_debounce_1 = require("throttle-debounce");
const log = Logger_1.Logger.child({
    namespace: 'subscribe',
});
const subscribe = (client, trigger, signal) => {
    return new Promise((resolve, reject) => {
        signal === null || signal === void 0 ? void 0 : signal.addEventListener('abort', () => {
            resolve(null);
        }, {
            once: true,
        });
        client.command([
            'subscribe',
            trigger.watch,
            trigger.id,
            {
                expression: trigger.expression,
                fields: ['name', 'size', 'mtime_ms', 'exists', 'type'],
                relative_root: trigger.relativePath,
            },
        ], (error, response) => {
            if (error) {
                reject(error);
                return;
            }
            log.info('subscription %s established', response.subscribe);
        });
        let activeTask = null;
        let first = true;
        let handleSubscriptionEvent = async (event) => {
            var _a;
            if (event.files.length > 10) {
                log.trace({
                    files: event.files.slice(0, 10).map((file) => {
                        return file.name;
                    }),
                }, '%d files changed; showing first 10', event.files.length);
            }
            else {
                log.trace({
                    files: event.files.map((file) => {
                        return file.name;
                    }),
                }, '%d files changed', event.files.length);
            }
            let reportFirst = first;
            if (first) {
                reportFirst = true;
                first = false;
            }
            let controller = null;
            if (trigger.interruptible) {
                controller = new AbortController();
            }
            if (activeTask) {
                if (trigger.interruptible) {
                    log.warn('aborted task %s (%s)', trigger.name, activeTask.id);
                    if (!activeTask.abortController) {
                        throw new Error('Expected abort controller to be set');
                    }
                    activeTask.abortController.abort();
                    activeTask = null;
                }
                else {
                    log.warn('waiting for %s (%s) task to complete', trigger.name, activeTask.id);
                    if (activeTask.queued) {
                        return;
                    }
                    activeTask.queued = true;
                    try {
                        await activeTask.promise;
                    }
                    catch (_b) {
                        // nothing to do
                    }
                }
            }
            const taskId = (0, generateShortId_1.generateShortId)();
            const triggerSignal = (_a = controller === null || controller === void 0 ? void 0 : controller.signal) !== null && _a !== void 0 ? _a : null;
            const taskPromise = (0, p_retry_1.default)((attempt) => {
                var _a;
                return trigger.onChange({
                    attempt,
                    // eslint-disable-next-line @typescript-eslint/no-explicit-any
                    files: event.files.map((file) => {
                        return {
                            exists: file.exists,
                            mtime: file.mtime_ms,
                            name: node_path_1.default.join(event.root, file.name),
                            size: file.size,
                        };
                    }),
                    first: reportFirst,
                    signal: triggerSignal,
                    spawn: (0, createSpawn_1.createSpawn)(taskId, triggerSignal),
                    warning: (_a = event.warning) !== null && _a !== void 0 ? _a : null,
                });
            }, {
                ...trigger.retry,
                onFailedAttempt: ({ retriesLeft }) => {
                    if (retriesLeft > 0) {
                        log.warn('retrying task %s (%s)...', trigger.name, taskId);
                    }
                },
            })
                // eslint-disable-next-line promise/prefer-await-to-then
                .then(() => {
                if (taskId === (activeTask === null || activeTask === void 0 ? void 0 : activeTask.id)) {
                    log.trace('completed task %s (%s)', trigger.name, taskId);
                    activeTask = null;
                }
            })
                // eslint-disable-next-line promise/prefer-await-to-then
                .catch((error) => {
                reject(error);
            });
            // eslint-disable-next-line require-atomic-updates
            activeTask = {
                abortController: controller,
                id: taskId,
                promise: taskPromise,
                queued: false,
            };
            log.trace('started task %s (%s)', trigger.name, taskId);
        };
        if (trigger.debounce) {
            handleSubscriptionEvent = (0, throttle_debounce_1.debounce)(trigger.debounce.wait, handleSubscriptionEvent, {
                atBegin: trigger.debounce.leading,
            });
        }
        client.on('subscription', async (event) => {
            if (event.subscription !== trigger.id) {
                return;
            }
            handleSubscriptionEvent(event);
        });
    });
};
exports.subscribe = subscribe;
